# SPDX-FileCopyrightText: 2014 Kai Uwe Broulik <kde@privat.broulik.de>
# SPDX-FileCopyrightText: 2014 Martin Klapetek <mklapetek@kde.org>
# SPDX-License-Identifier: GPL-2.0-or-later

$XGETTEXT $(find . -name \*.cpp -o -name \*.h -o -name \*.qml) -o $podir/kcm_mobile_time.pot
